/*
	Copyright (c) 2001 Maurice Gittens <maurce@gittens.nl>

   This software is distributed under the GNU General
   Public License.
	See the file COPYING for more information on usage and
	redistribution of this file and for a DISCLAIMER OF ALL
	WARRANTIES.

	Written by Maurice Gittens <maurice@gittens.nl>.
*/

#include <tiny_parser.h>
#include <tiny_lexer.h>
#include <tiny_principle.h>
#include <tiny_action.h>
#include <tiny_rewritablesymbol.h>
#include <tiny_genericsymbol.h>
#include <iostream>

using namespace std;

typedef tiny_lexer Lex;

void tiny_parser::operator()()
{
   advance();
   while(!lexer.eof())
      fileDeclarationList();

   if (!lexer.getErrorCount())
   {
      if (!lexer.getStartSymbol())
         lexer.setStartSymbol(lexer.getNonTerminal(0).getId());
      if (lexer.getTypeName().empty())
         lexer.setTypeName("int");
   }
}

void tiny_parser::fileDeclarationList()
{
	do
	{
	   fileDeclaration();
      advance();
	}
   while (lookAhead);
}

void tiny_parser::fileDeclaration()
{
   switch(lookAhead)
	{
      case Lex::RW_GRAMMAR :
         grammarHead();

         if (!currentIs(Lex::LC))
         {
            endFileDeclaration();
            return;
         }
         lexer.setIsGlobal(false);
         grammarDeclSectionList();

         principlesSection();

         if (lookAhead != Lex::RC)
         {
            endFileDeclaration();
            return;
         }
         lexer.setIsGlobal(true);
      return;
      case Lex::RW_TERMINALS :
         symbolDeclSection(tiny_grammarSymbol::terminal);
      break;
      case Lex::RW_GENERICS :
           lexer.fatalError("global generic nonterminals not implemented yet");
      return;
      case Lex::RW_HEADER :
      {
         nextIs(Lex::CODEBLOCK);
         std::string cb = lexer.getHeaderCodeBlock();
         cb += lexer.getCodeBlock();
         lexer.setHeaderCodeBlock(cb);
         return;
      }
      case Lex::RW_SOURCE :
      {
         nextIs(Lex::CODEBLOCK);
         std::string cb = lexer.getSourceCodeBlock();
         cb += lexer.getCodeBlock();
         lexer.setSourceCodeBlock(cb);
         return;	
      }
	   default:
	      lexer.error(std::string("syntax error: read '") +
	                  lexer.getTokenClass(lookAhead)
		               + "' when grammar, header, source or terminals"
				   		" was expected");
   	
         endFileDeclaration();
	}
}

void tiny_parser::grammarHead()
{
   if (!nextIs(Lex::ID))
   {
      endFileDeclaration();
      return;
   }

   lexer.setGrammarName(lexer.getLexeme());

   advance();

   if (lookAhead == Lex::CODEBLOCK)
   {
      lexer.setInheritBlock(lexer.getCodeBlock());
      advance();
   }

   lexer.setPlaybackSymbol();
   lexer.setAnySymbol();
}

void tiny_parser::grammarDeclSectionList()
{
	do
	{
      grammarDeclSection();
	}
   while (lookAhead == Lex::RW_TERMINALS ||
          lookAhead == Lex::RW_NONTERMINALS ||
          lookAhead == Lex::RW_START ||
          lookAhead == Lex::RW_ERROR ||
          lookAhead == Lex::RW_GENERICS ||
          lookAhead == Lex::RW_TYPE ||
          lookAhead == Lex::CODEBLOCK);
}

void tiny_parser::grammarDeclSection()
{
      switch(lookAhead)
	   {
         case Lex::RW_TERMINALS :
            symbolDeclSection(tiny_grammarSymbol::terminal);
         return;
         case Lex::RW_NONTERMINALS :
            symbolDeclSection(tiny_grammarSymbol::nonterminal);
         return;
         case Lex::RW_GENERICS :
            symbolDeclSection(tiny_grammarSymbol::generic);
         return;
         case Lex::RW_START :
         {
            if (!nextIs(Lex::NONTERMINAL))
            {
               endGrammarDeclaration();
               return;
            }

            if (!lexer.getStartSymbol())
            {
               tiny_grammarSymbol* sym;
               sym = lexer.getSymbol(lexer.getLexeme());

               lexer.setStartSymbol(sym->getEncodedId());
            }
            else
            {
               lexer.error("start symbol may not be defined more than once");
            }

            advance();
         }
         return;
         case Lex::RW_ERROR :
         {
            if (!nextIs(Lex::NONTERMINAL))
            {
               endGrammarDeclaration();
               return;
            }

            if (!lexer.getErrorSymbol())
            {
               tiny_grammarSymbol* sym;
               sym = lexer.getSymbol(lexer.getLexeme());

               lexer.setErrorSymbol(sym->getEncodedId());
            }
            else
            {
               lexer.error("error symbol may not be defined more than once");
            }
            advance();
         }
         return;
         case Lex::RW_TYPE :
         {
            if (!nextIs(Lex::ID))
            {
               endGrammarDeclaration();
               return;
            }

            if (lexer.getTypeName().empty())
               lexer.setTypeName(lexer.getLexeme());
            else
               lexer.error("typename may not be defined more than once");

            advance();
         }
         return;
         case Lex::CODEBLOCK :
         {
            std::string cb = lexer.getHeaderCodeBlock();
            cb += lexer.getCodeBlock();
            lexer.setHeaderCodeBlock(cb);
            advance();
         return;
         }
	      default:
	         lexer.error(std::string("syntax error: read '") +
	                  lexer.getTokenClass(lookAhead)
		               + "' when grammar, header, source or terminals"
				   		" was expected");
   	
            endGrammarDeclaration();
   	}
}

void tiny_parser::principlesSection()
{
   if(!currentIs(Lex::RW_PRINCIPLES))
   {
      endPrinciplesSection();
      return;
   }

	do
	{
      if (lookAhead == Lex::CODEBLOCK)
      {
         std::string cb = lexer.getHeaderCodeBlock();
         cb += lexer.getCodeBlock();
         lexer.setHeaderCodeBlock(cb);
         advance();
      }
      else if (lookAhead == Lex::RC)
      {
         break;
      }
      else if (lookAhead == Lex::GENERIC)
      {
         genericSymbolDefinition();
      }
      else
         symbolDefinition();

	}
   while (lookAhead);
}

void tiny_parser::symbolDefinition()
{
   tiny_principle principle;
   tiny_rewritableSymbol* nonterminal = 0;

   if (!symbolDefLhs(principle, &nonterminal))
   {
      endSymbolDefinition();
      return;
   }

   if (lookAhead == Lex::COLON)
   {
      advance();
      principleRhsList(principle);
      if (nonterminal)
         nonterminal->addRule(principle);
   }
   else
   {
      endSymbolDefinition();
   }
}

bool tiny_parser::symbolDefLhs(tiny_principle& principle,
                               tiny_rewritableSymbol** nt,
                               bool isGeneric)
{
   symbolList(principle.getLHS());

   principle.getLHS().push_back(0);

   if (!currentIs(Lex::DOT))
   {
      return false;
   }

   tiny_grammarSymbol* sym;
   sym = lexer.getSymbol(lexer.getLexeme());

   if (!sym)
   {
         lexer.newNonTerminal();
         lookAhead = Lex::NONTERMINAL;
         lexer.error(string("undeclared symbol: '")
                     + lexer.getLexeme() + "'");
         sym = lexer.getSymbol(lexer.getLexeme());
   }
   else
   {
      principle.getLHS().push_back(sym->getEncodedId());
      sym->setUsed();
   }

   if (lookAhead == Lex::TERMINAL)
   {
      lexer.error("terminals not allowed here");
   } else if ( isGeneric && !currentIs(Lex::GENERIC))
   {
      return false;
   }
   else if ( !isGeneric && !currentIs(Lex::NONTERMINAL))
   {
      return false;
   }

   tiny_rewritableSymbol& nonterminal =
      dynamic_cast<tiny_rewritableSymbol&>(*sym);

   symbolList(principle.getLHS());

   if (lookAhead == Lex::CODEBLOCK)
   {
      tiny_action& act = lexer.newAction(lexer.getCodeBlock());
      principle.getLHS().push_back(act.getEncodedId());
      advance();
   }

   *nt = &nonterminal;

   return true;
}

bool tiny_parser::parmList()
{
   if (!currentIs(Lex::LP))
   {
      return false;
   }

   if (lookAhead == Lex::RP)
   {
      advance();
      return true;
   }

   for(;;)
   {
      if (lookAhead == Lex::ID ||
          lookAhead == Lex::TERMINAL ||
          lookAhead == Lex::GENERIC ||
          lookAhead == Lex::NONTERMINAL)
      {
         lexer.newVariable(lexer.getLexeme());
         advance();

         if (lookAhead == Lex::COMMA)
         {
            advance();
            continue;
         }

         if (lookAhead == Lex::RP)
         {
            advance();
         }
         else
         {
	         lexer.error(std::string("syntax error: read '") +
	                  lexer.getTokenClass(lookAhead)
		               + "' when ',' or ')' was expected");
            return false;
         }
         return true;
      }
      else
      {
	      lexer.error(std::string("syntax error: read '") +
	                  lexer.getTokenClass(lookAhead)
		               + "' when parmetername was expected");
         return false;
      }
   }
}

void tiny_parser::genericSymbolDefinition()
{
   lexer.clearVariables();

   tiny_genericSymbol& sym =
      dynamic_cast<tiny_genericSymbol&>(*lexer.getSymbol(lexer.getLexeme()));
   advance();

   if (!parmList())
   {
      endSymbolDefinition();
      return;
   }

   if (!currentIs(Lex::LC))
   {
      endSymbolDefinition();
      return;
   }

   do
   {
      tiny_principle principle;
      tiny_rewritableSymbol* nonterminal = 0;

      if (!symbolDefLhs(principle, &nonterminal, true))
      {
         endSymbolDefinition();
         return;
      }

      if (sym.getName() != nonterminal->getName())
      {
	      lexer.error(sym.getName() + " not used next to dot");
      }

      if (lookAhead == Lex::COLON)
      {
         advance();
         principleRhsList(principle);
         nonterminal->addRule(principle);
      }
      else
      {
         endSymbolDefinition();
      }
   }
   while(lookAhead && lookAhead != Lex::RC);

   if (!currentIs(Lex::RC))
   {
         endSymbolDefinition();
   }

   lexer.clearVariables();
}

void tiny_parser::principleRhsList(tiny_principle& principle)
{
   for(;;)
   {
      principleRhs(principle);

      if (lookAhead == Lex::BAR)
      {
         advance();
      }
      else if (lookAhead == Lex::SEMIC)
      {
         advance();
         break;
      }
      else
      {
        lexer.error(string("';' or '|' expected; found ") +
               lexer.getTokenClass(lookAhead) + " instead");
      }

   }
}

void tiny_parser::principleRhs(tiny_principle& principle)
{
   tiny_principle::RightHandSide& rhs = principle.newRHS();
   if (lookAhead == Lex::CODEBLOCK)
   {
      tiny_action& act = lexer.newAction(lexer.getCodeBlock());
      rhs.getString().push_back(act.getEncodedId());
      advance();
   }

   symbolList(rhs.getString());
   if (!currentIs(Lex::DOT))
   {
      endSymbolDefinition();
   }
   rhs.push_back(0);
   symbolList(rhs.getString());

   if (lookAhead == Lex::RECORD)
   {
      advance();
      symbolList(rhs.getStringToRecord());
      if (!currentIs(Lex::DOT))
      {
         endSymbolDefinition();
      }
      rhs.getStringToRecord().push_back(0);
      symbolList(rhs.getStringToRecord());
   }

   if (lookAhead == Lex::CODEBLOCK)
   {
      tiny_action& act = lexer.newAction(lexer.getCodeBlock());
      rhs.getString().push_back(act.getEncodedId());
      advance();
   }

}

void tiny_parser::symbolList(tiny_symbolString &ss)
{
   while(lookAhead == Lex::NONTERMINAL ||
         lookAhead == Lex::TERMINAL ||
         lookAhead == Lex::VARIABLE ||
         lookAhead == Lex::GENERIC ||
         lookAhead == Lex::PLAYBACK ||
         lookAhead == Lex::ID)
   {
      if (lookAhead == Lex::ID)
      {
         lexer.newNonTerminal();
         lookAhead = Lex::NONTERMINAL;
         lexer.error(string("undeclared symbol: '")
                     + lexer.getLexeme() + "'");
      } else if (lookAhead == Lex::PLAYBACK)
      {
         ss.push_back(lexer.getPlaybackSymbol());
         advance();
         continue;
      }

      tiny_grammarSymbol* sym;

      if (lookAhead != Lex::VARIABLE)
         assert( sym = lexer.getSymbol(lexer.getLexeme()) );
      else
         assert( sym = lexer.getSymbol((string("$")+lexer.getLexeme()).c_str()) );

      ss.push_back(sym->getEncodedId());
      sym->setUsed();
      advance();
   }

}

void tiny_parser::symbolDeclSection(tiny_grammarSymbol::Type type)
{
   advance();
   while(lookAhead == Lex::ID ||
         lookAhead == Lex::NONTERMINAL ||
         lookAhead == Lex::TERMINAL)
   {
      if (lookAhead == Lex::ID)
      {
         switch(type)
         {
            case tiny_grammarSymbol::terminal :
               lexer.newTerminal();
            break;
            case tiny_grammarSymbol::nonterminal :
               lexer.newNonTerminal();
            break;
            case tiny_grammarSymbol::generic :
               lexer.newGeneric();
            break;
            default:
               lexer.fatalError("bad symbol type; probably a bug");
            break;
         }
      }
      else
         lexer.error(string("symbol '") + lexer.getLexeme() +
                     "' allready declared");

      advance();


   }
}

void tiny_parser::endSymbolDefinition()
{
   while(lookAhead)
   {
      advance();
      if (lookAhead == Lex::SEMIC)
         break;
   }

}

void tiny_parser::endPrinciplesSection()
{
   endFileDeclaration();
}

void tiny_parser::endGrammarRules()
{
   endFileDeclaration();
}

void tiny_parser::endGrammarDeclaration()
{
   endFileDeclaration();
}

void tiny_parser::endFileDeclaration()
{
   while(!lexer.eof())
	{
      if (lookAhead == Lex::RW_HEADER    ||
		    lookAhead == Lex::RW_SOURCE    ||
		    lookAhead == Lex::RW_TERMINALS ||
		    lookAhead == Lex::RW_GRAMMAR)
      {
		   break;
      }
      advance();
   }
}

void tiny_parser::endGenericSymbolDefinition()
{
   while(!lexer.eof())
	{
      if (lookAhead == Lex::RC)
      {
		   break;
      }
      advance();
   }
}
