/*
	Copyright (c) 1996 Gittens Information Systems B.V.

	See the file license.terms for information on usage and
	redistribution of this file and for a DISCLAIMER OF ALL
	WARRANTIES.

*/

#ifndef TINYSTACK_SEEN
#define TINYSTACK_SEEN

#include <tiny_exception.h>

const int TINY_MAX_RHS_LEN 			= 32;
const int TINY_DEFAULT_STACK_SIZE		= 1024;

template <class STACKTYPE>
class TINY_Stack
{
public:
	TINY_Stack(unsigned size, unsigned int maxRhsLen = TINY_MAX_RHS_LEN);
	void push(STACKTYPE);
	STACKTYPE pop();
	const STACKTYPE& peek(int) const;
	const STACKTYPE *peekAddress(int) const;
	const STACKTYPE &access(int) const;
	STACKTYPE *peekAddress(int);
	STACKTYPE &access(int);
	bool empty() const;
	int itemCount() const;
	void pop(int);
	void reset();
	~TINY_Stack();

protected:
	unsigned size;
	unsigned maxRhsLen;
	STACKTYPE *buffer;
	STACKTYPE *bottom;
	STACKTYPE *top;
	STACKTYPE *sp1;

private:
   TINY_Stack(const TINY_Stack&);
   TINY_Stack& operator=(const TINY_Stack&);

   int pushOk() const;
	int popOk() const;
};

template <class STACKTYPE>
class TINY_DoubleStack : protected TINY_Stack<STACKTYPE>
{
public:
	TINY_DoubleStack(unsigned size = TINY_DEFAULT_STACK_SIZE, const unsigned int maxRhsLen = MAX_RHS_LEN);

	void pushL(STACKTYPE);
	void pushR(STACKTYPE);

	STACKTYPE popL();
	STACKTYPE popR();

	void popL(int);
	void popR(int);

	int itemCountL() const;
	int itemCountR() const;
	int itemCount() const;

	const STACKTYPE& peekL(int) const;
	const STACKTYPE& peekR(int) const;
	const STACKTYPE *peekAddressL(int) const;
	const STACKTYPE *peekAddressR(int) const;
	STACKTYPE *peekAddressL(int);
	STACKTYPE *peekAddressR(int);

	const STACKTYPE &accessL(int) const;
	const STACKTYPE &accessR(int) const;
	STACKTYPE &accessL(int);
	STACKTYPE &accessR(int);

	void reset();

	bool empty() const;
	bool emptyR() const;
	bool emptyL() const;

	~TINY_DoubleStack();
private:
   TINY_DoubleStack(const TINY_DoubleStack&);
   TINY_DoubleStack& operator=(const TINY_DoubleStack&);
	int pushOk() const;
	int popLOk() const;
	int popROk() const;

	STACKTYPE *sp2;
};

template <class STACKTYPE>
class TINY_ContextMachine
{
public:
   TINY_ContextMachine(unsigned size = TINY_DEFAULT_STACK_SIZE,
                      const unsigned int maxRhsLen = TINY_MAX_RHS_LEN)
                        : s1(size, maxRhsLen), s2(size, maxRhsLen)
   {
      current = &s1;
   }

   const TINY_DoubleStack<STACKTYPE>* operator->() const { return current; }
   TINY_DoubleStack<STACKTYPE>* operator->() { return current; }

   const TINY_DoubleStack<STACKTYPE>& operator*() const { return *current; }
   TINY_DoubleStack<STACKTYPE>& operator*() { return *current; }

   void swap() { current = (current == &s1 ? &s2 : &s1); }
   void operator~() { return swap(); }
private:
   TINY_ContextMachine(const TINY_ContextMachine&);
   TINY_ContextMachine& operator=(const TINY_ContextMachine&);

   TINY_DoubleStack<STACKTYPE> s1, s2 , *current;
};


/* The stack methods */
/*-------------------*/

//
// The reset method
//
template<class T>
void TINY_Stack<T>::reset()
{
	bottom = buffer + maxRhsLen;
	top = sp1 = bottom+size;
}

//
// The pop(int) method
//
template<class T>
void TINY_Stack<T>::pop(int i)
{
	sp1 += i;
}

//
// The constructor
//
template<class T>
TINY_Stack<T>::TINY_Stack(unsigned s, unsigned gs)
{
	buffer = new T[s + 2 * (maxRhsLen = gs)];

	size = s;

	bottom =  buffer + maxRhsLen;

	top = sp1 = bottom + s;
}

//
// The member function empty
//
template<class T>
bool TINY_Stack<T>::empty() const
{
	return sp1 == top;
}

//
// The memberfunction pushOk();
//
template <class T>
int TINY_Stack<T>::pushOk() const
{
	return sp1 > bottom ? 1 : 0;
}

//
// The memberfunction popOk
//
template <class T>
int TINY_Stack<T>::popOk() const
{
	return sp1 < top;
}

//
// The member function push()
//
template<class T>
void TINY_Stack<T>::push(T v)
{
	if (pushOk())
	{
		*--sp1 = v;
		return;
	}

	throw tiny_exception("stack overflow");
}

//
// The member function pop()
//
template <class T>
T TINY_Stack<T>::pop()
{
	if (popOk())
		return *sp1++;

	throw tiny_exception("stack underflow");
}

//
// The member function peek(int)
//
template <class T>
const T& TINY_Stack<T>::peek(int i) const
{
	return *(sp1+i);
}

//
// The member function peek(int)
//
template <class T>
T *TINY_Stack<T>::peekAddress(int i)
{
	return sp1+i;
}

//
// The member function peek(int)
//
template <class T>
const T *TINY_Stack<T>::peekAddress(int i) const
{
	return sp1+i;
}

//
// The member function access(int)
//
template <class T>
T &TINY_Stack<T>::access(int i)
{
	return *(sp1+i);
}

//
// The member function access(int)
//
template <class T>
const T &TINY_Stack<T>::access(int i) const
{
	return *(sp1+i);
}

//
//
//
template <class T>
TINY_Stack<T>::~TINY_Stack()
{
	delete [] buffer;
}

//
//
//
template<class T>
int TINY_Stack<T>::itemCount() const
{
	return top - sp1;
}

/* DoubleStack Methods */
/*---------------------*/

//
// The reset method
//
template<class T>
void TINY_DoubleStack<T>::reset()
{
	TINY_Stack<T>::reset();
	sp2 = bottom;
}


//
// The popL(int) method
//
template<class T>
void TINY_DoubleStack<T>::popL(int i)
{
	TINY_Stack<T>::pop(i);
}

//
// The popR(int) method
//
template<class T>
void TINY_DoubleStack<T>::popR(int i)
{
	sp2 -= i;
}

//
// The constructor
//
template <class T>
TINY_DoubleStack<T>::TINY_DoubleStack(unsigned int s, const unsigned int maxRhsLen) : TINY_Stack<T>(s, maxRhsLen)
{
	sp2 = bottom;
}

//
// The empty method
//
template <class T>
bool TINY_DoubleStack<T>::empty() const
{
	return TINY_Stack<T>::empty() && sp2 == bottom;
}

//
//
//
template <class T>
bool TINY_DoubleStack<T>::emptyR() const
{
	return sp2 == bottom;
}

//
//
//
template <class T>
bool TINY_DoubleStack<T>::emptyL() const
{
	return TINY_Stack<T>::empty();
}

//
// The pushOk method
//
template <class T>
int TINY_DoubleStack<T>::pushOk() const
{
	return (sp1 - sp2) > 0;
}

//
// The popLOk method
//
template<class T>
int TINY_DoubleStack<T>::popLOk() const
{
	return sp1 < top;
}

//
// The popROk method
//
template<class T>
int TINY_DoubleStack<T>::popROk() const
{
	return sp2 > bottom;
}

//
//
//
template <class T>
void TINY_DoubleStack<T>::pushL(T v)
{
	if (pushOk())
	{
		*--sp1 = v;
		return;
	}

	throw tiny_exception("stack overflow");
}

//
//
//
template <class T>
void TINY_DoubleStack<T>::pushR(T v)
{
	if(pushOk())
	{
		*sp2++ = v;
		return;
	}

	throw tiny_exception("stack overflow");
}

//
//
//
template<class T>
T TINY_DoubleStack<T>::popL()
{
	if (popLOk())
		return *sp1++;

	throw tiny_exception("stack underflow");
}

//
//
//
template<class T>
T TINY_DoubleStack<T>::popR()
{
	if (popROk())
		return *--sp2;

	throw tiny_exception("stack underflow");
}

//
//
//
template<class T>
const T& TINY_DoubleStack<T>::peekL(int i) const
{
	return peek(i);
}

//
//
//
template<class T>
T *TINY_DoubleStack<T>::peekAddressL(int i)
{
	return peekAddressL(i);
}

//
//
//
template<class T>
const T *TINY_DoubleStack<T>::peekAddressL(int i) const
{
	return peekAddressL(i);
}

//
//
//
template<class T>
T &TINY_DoubleStack<T>::accessL(int i)
{
	return access(i);
}

//
//
//
template<class T>
const T &TINY_DoubleStack<T>::accessL(int i) const
{
	return access(i);
}

//
//
//
template<class T>
int TINY_DoubleStack<T>::itemCountL() const
{
	return TINY_Stack<T>::itemCount();
}

//
//
//
template <class T>
const T& TINY_DoubleStack<T>::peekR(int i) const
{
	return *(sp2-i-1);
}

//
//
//
template <class T>
T *TINY_DoubleStack<T>::peekAddressR(int i)
{
	return sp2-i-1;
}

//
//
//
template <class T>
const T *TINY_DoubleStack<T>::peekAddressR(int i) const
{
	return sp2-i-1;
}

//
//
//
template <class T>
T &TINY_DoubleStack<T>::accessR(int i)
{
	return *(sp2-i-1);
}

//
//
//
template <class T>
const T &TINY_DoubleStack<T>::accessR(int i) const
{
	return *(sp2-i-1);
}

//
//
//
template<class T>
int TINY_DoubleStack<T>::itemCountR() const
{
	return sp2 - bottom;
}

//
//
//
template<class T>
int TINY_DoubleStack<T>::itemCount() const
{
	return itemCountR() + itemCountL();
}

//
//
//
template<class T>
TINY_DoubleStack<T>::~TINY_DoubleStack()
{
}

#endif
