/*
	Copyright (c) 2001 Maurice Gittens <maurce@gittens.nl>

   This software is distributed under the GNU General
   Public License.
	See the file COPYING for more information on usage and
	redistribution of this file and for a DISCLAIMER OF ALL
	WARRANTIES.

	Written by Maurice Gittens <maurice@gittens.nl>.
*/

#ifndef TINY_STACKUTILS_SEEN
#define TINY_STACKUTILS_SEEN

#include <tiny_exception.h>

const int TINY_MAX_RHS_LEN 			   = 16;
const int TINY_DEFAULT_STACK_SIZE		= 128;
const int TINY_GUARDSIZE	 			   = 16;

template <class STACKTYPE>
class tiny_Stack
{
public:
	tiny_Stack(unsigned size, unsigned int guardSize = TINY_MAX_RHS_LEN);
	void push(STACKTYPE);
	STACKTYPE pop();
	STACKTYPE peek(int);
	STACKTYPE *peekAddress(int);
	STACKTYPE &access(int);
	int Empty() const;
	int itemCount() const;
	void pop(int);
	void reset();
	~tiny_Stack();

protected:
	unsigned size;
	unsigned guardSize;
	STACKTYPE *buffer;
	STACKTYPE *bottom;
	STACKTYPE *top;
	STACKTYPE *sp1;

private:
	int pushOk();
	int popOk();
};

template <class STACKTYPE>
class tiny_DoubleStack : protected tiny_Stack<STACKTYPE>
{
public:
	tiny_DoubleStack(unsigned size = TINY_DEFAULT_STACK_SIZE, const unsigned int guardSize = TINY_GUARDSIZE);

	void pushL(STACKTYPE);
	void pushR(STACKTYPE);

	STACKTYPE popL();
	STACKTYPE popR();

	void popL(int);
	void popR(int);

	int itemCountL() const;
	int itemCountR() const;
	int itemCount() const;

	STACKTYPE peekL(int);
	STACKTYPE peekR(int);
	STACKTYPE *peekAddressL(int);
	STACKTYPE *peekAddressR(int);

	STACKTYPE &accessL(int);
	STACKTYPE &accessR(int);

	void reset();

	int Empty() const;
	int EmptyR() const;
	int EmptyL() const;

	~tiny_DoubleStack();
protected:
	int pushOk();
	int popLOk();
	int popROk();

	STACKTYPE *sp2;
};

template <class STACKTYPE>
class tiny_ContextMachine
{
public:
   tiny_ContextMachine(unsigned size = TINY_DEFAULT_STACK_SIZE, const unsigned int guardSize = TINY_GUARDSIZE) : s1(size, guardSize), s2(size, guardSize)
   {
      current = &s1;
   }

   tiny_DoubleStack<STACKTYPE>* operator->() { return current; }

   tiny_DoubleStack<STACKTYPE>& Switch() { return current == &s1 ? s2 : s1; }
private:

   tiny_DoubleStack<STACKTYPE> s1, s2 , *current;
};


/* The stack methods */
/*-------------------*/

//
// The reset method
//
template<class T>
void tiny_Stack<T>::reset()
{
	bottom = buffer + guardSize;
	top = sp1 = bottom+size;

}

//
// The pop(int) method
//
template<class T>
void tiny_Stack<T>::pop(int i)
{
	sp1 += i;
}

//
// The constructor
//
template<class T>
tiny_Stack<T>::tiny_Stack(unsigned s, unsigned gs)
{
	buffer = new T[s + 2 * (guardSize = gs)];

	size = s;

	bottom =  buffer + guardSize;

	top = sp1 = bottom + s;
}

//
// The member function Empty
//
template<class T>
int tiny_Stack<T>::Empty() const
{
	return sp1 == top;
}

//
// The memberfunction pushOk();
//
template <class T>
int tiny_Stack<T>::pushOk()
{
	return sp1 > bottom ? 1 : 0;
}

//
// The memberfunction popOk
//
template <class T>
int tiny_Stack<T>::popOk()
{
	return sp1 < top;
}

//
// The member function push()
//
template<class T>
void tiny_Stack<T>::push(T v)
{
	if (pushOk())
	{
		*--sp1 = v;
		return;
	}

	throw tiny_exception("stack overflow");
}

//
// The member function pop()
//
template <class T>
T tiny_Stack<T>::pop()
{
	if (popOk())
		return *sp1++;

	throw tiny_exception("stack underflow");
}

//
// The member function peek(int)
//
template <class T>
T tiny_Stack<T>::peek(int i)
{
	return *(sp1+i);
}

//
// The member function peek(int)
//
template <class T>
T *tiny_Stack<T>::peekAddress(int i)
{
	return sp1+i;
}

//
// The member function access(int)
//
template <class T>
T &tiny_Stack<T>::access(int i)
{
	return *(sp1+i);
}

//
//
//
template <class T>
tiny_Stack<T>::~tiny_Stack()
{
	delete [] buffer;
}

//
//
//
template<class T>
int tiny_Stack<T>::itemCount() const
{
	return top - sp1;
}

/* DoubleStack Methods */
/*---------------------*/

//
// The reset method
//
template<class T>
void tiny_DoubleStack<T>::reset()
{
	tiny_Stack<T>::reset();
	sp2 = bottom;
}


//
// The popL(int) method
//
template<class T>
void tiny_DoubleStack<T>::popL(int i)
{
	tiny_Stack<T>::pop(i);
}

//
// The popR(int) method
//
template<class T>
void tiny_DoubleStack<T>::popR(int i)
{
	sp2 -= i;
}

//
// The constructor
//
template <class T>
tiny_DoubleStack<T>::tiny_DoubleStack(unsigned int s, const unsigned int guardSize) : tiny_Stack<T>(s, guardSize)
{
	sp2 = bottom;
}

//
// The Empty method
//
template <class T>
int tiny_DoubleStack<T>::Empty() const
{
	return tiny_Stack<T>::Empty() && sp2 == bottom;
}

//
//
//
template <class T>
int tiny_DoubleStack<T>::EmptyR() const
{
	return sp2 == bottom;
}

//
//
//
template <class T>
int tiny_DoubleStack<T>::EmptyL() const
{
	return tiny_Stack<T>::Empty();
}

//
// The pushOk method
//
template <class T>
int tiny_DoubleStack<T>::pushOk()
{
	return (sp1 - sp2) > 0;
}

//
// The popLOk method
//
template<class T>
int tiny_DoubleStack<T>::popLOk()
{
	return sp1 < top;
}

//
// The popROk method
//
template<class T>
int tiny_DoubleStack<T>::popROk()
{
	return sp2 > bottom;
}

//
//
//
template <class T>
void tiny_DoubleStack<T>::pushL(T v)
{
	if (pushOk())
	{
		*--sp1 = v;
		return;
	}

	throw tiny_exception("stack overflow");
}

//
//
//
template <class T>
void tiny_DoubleStack<T>::pushR(T v)
{
	if(pushOk())
	{
		*sp2++ = v;
		return;
	}

	throw tiny_exception("stack overflow");
}

//
//
//
template<class T>
T tiny_DoubleStack<T>::popL()
{
	if (popLOk())
		return *sp1++;

	throw tiny_exception("stack underflow");
}

//
//
//
template<class T>
T tiny_DoubleStack<T>::popR()
{
	if (popROk())
		return *--sp2;

	throw tiny_exception("stack underflow");
}

//
//
//
template<class T>
T tiny_DoubleStack<T>::peekL(int i)
{
	return peek(i);
}

//
//
//
template<class T>
T *tiny_DoubleStack<T>::peekAddressL(int i)
{
	return peekAddressL(i);
}

//
//
//
template<class T>
T &tiny_DoubleStack<T>::accessL(int i)
{
	return access(i);
}

//
//
//
template<class T>
int tiny_DoubleStack<T>::itemCountL() const
{
	return tiny_Stack<T>::itemCount();
}

//
//
//
template <class T>
T tiny_DoubleStack<T>::peekR(int i)
{
	return *(sp2-i-1);
}

//
//
//
template <class T>
T *tiny_DoubleStack<T>::peekAddressR(int i)
{
	return sp2-i-1;
}

//
//
//
template <class T>
T &tiny_DoubleStack<T>::accessR(int i)
{
	return *(sp2-i-1);
}

//
//
//
template<class T>
int tiny_DoubleStack<T>::itemCountR() const
{
	return sp2 - bottom;
}

//
//
//
template<class T>
int tiny_DoubleStack<T>::itemCount() const
{
	return itemCountR() + itemCountL();
}

//
//
//
template<class T>
tiny_DoubleStack<T>::~tiny_DoubleStack()
{
}

#endif
