/*
	Copyright (c) 2001 Maurice Gittens <maurce@gittens.nl>

   This software is distributed under the GNU General
   Public License.
	See the file COPYING for more information on usage and
	redistribution of this file and for a DISCLAIMER OF ALL
	WARRANTIES.

	Written by Maurice Gittens <maurice@gittens.nl>.
*/

#include <tiny_symbolprocessor.h>
#include <tiny_rewritablesymbol.h>
#include <tiny_util.h>
#include <tiny_lexer.h>
#include <tiny_action.h>
#include <iostream>
#include <algorithm>
#include <cstdio>

//#define DEBUG
#ifdef DEBUG
#define D(x) x
#else
#define D(x)
#endif

using namespace std;

void tiny_symbolProcessor::traverse(void (tiny_symbolProcessor::*f)(tiny_action&))
{
   for(unsigned i = 0; i < scanner.getActionCount(); i++)
      (this ->* f)(scanner.getAction(i));
}

void tiny_symbolProcessor::traverse(void (tiny_symbolProcessor::*f)(tiny_rewritableSymbol&))
{
	for(unsigned i = 0; i < scanner.getNonTerminalCount(); i++)
		(this ->* f)(scanner.getNonTerminal(i));
}

//
//
//
void tiny_symbolProcessor::checkUsage()
{
   if (checkTerminals)
	{
      const int tcount = scanner.getTerminalCount();
      int i;
      for(i = 0; i < tcount; i++)
		{
			if (!scanner.getTerminal(i).isUsed())
			{
				char buf[32];
				sprintf(buf,"%d", scanner.getTerminal(i).getLineNo());
				scanner.warning("The terminal " + scanner.getTerminal(i).getName() +
									 " not used; line: " + buf);
      	}
      }
	}

#if 1	
	for(unsigned i = 0; i < scanner.getNonTerminalCount(); i++)
	{
		if (!scanner.getNonTerminal(i).isUsed())
		{
			char buf[32];
			sprintf(buf,"%d", scanner.getNonTerminal(i).getLineNo());
			scanner.warning("The non terminal " + scanner.getNonTerminal(i).getName() +
								 " not used; line: " + buf);
	
	   }
	}
#endif
}

//
//
//
void tiny_symbolProcessor::doSortSymbolRules()
{
	traverse(&tiny_symbolProcessor::sortSymbolRules);
}

//
//
//
void tiny_symbolProcessor::doReserveMemoryForFirstSets()
{
	traverse(&tiny_symbolProcessor::reserveMemoryForFirstSets);
}

//
//
//
void tiny_symbolProcessor::reserveMemoryForFirstSets(tiny_rewritableSymbol& nt)
{
	nt.setFirstSetCapacity(scanner.getTerminalCount());	
}

//
//
//
void tiny_symbolProcessor::doComputeConnections()
{
	traverse(&tiny_symbolProcessor::computeConnections);
}

//
//
//
void tiny_symbolProcessor::computeConnections(tiny_rewritableSymbol& nt)
{
	const int ruleCount = nt.getRuleCount();
	
	for(int i = 0; i < ruleCount; i++)
	{
		tiny_principle &rule = nt.getRule(i);
		const int altCount = rule.getAlternativeCount();

  		for(int j = 0 ; j < altCount; j++)
   	{
	   	tiny_principle::RightHandSide &rhs = rule.getRHS(j);

		   connectRHS(rhs);
  		}
	}

}

//
//
//
void tiny_symbolProcessor::connectRHS(tiny_principle::RightHandSide& rhs)
{
	tiny_symbolString string = rhs.getString();

	if( rhs.isNullable() || isAction(string.getSymbol(0)))
	{
		return;
	}

	int index = getDotIndex(&string);
	const int length = rhs.length();

	while(index < (length-1) && isTerminal(string.getSymbol(index + 1)))
	{
		string.advanceDot(index++);
	}

	++index;
	if (index < length && isNonTerminal(string.getSymbol(index)) )
	{
		tiny_rewritableSymbol& nt = scanner.getNonTerminal(makeId(string.getSymbol(index)));
		const int ruleCount = nt.getRuleCount();
	   tiny_principle *bestLhs = 0;

		for(int i = 0; i < ruleCount; ++i)
		{
			tiny_principle &rule = nt.getRule(i);
			if (rule.getLHS().match(string))
			{
			   if (!bestLhs)
			      bestLhs = &rule;
			   else
			   {
			      if (bestLhs->getLHS().length() < rule.getLHS().length())
				      bestLhs = &rule;
				}
			}
		}
		
		if (bestLhs)
			rhs.setRule(bestLhs);
		
   }
}

//
//
//
void tiny_symbolProcessor::doComputeSelectionSets()
{
	traverse(&tiny_symbolProcessor::computeSelectionSets);
}

//
//
//
void tiny_symbolProcessor::computeSelectionSets(tiny_rewritableSymbol& nt)
{
	const int ruleCount = nt.getRuleCount();
	for(int i = 0; i < ruleCount; i++)
	{
		tiny_principle &rule = nt.getRule(i);

		if (rule.isSingleton() && !rule.getRHS(0).isLeaf())
		{
			rule.getFirstSet()[makeId(scanner.getAnySymbol())] = true;
			rule.getRHS(0).getFirstSet()[makeId(scanner.getAnySymbol())] = true;
		}
		else if (!rule.isSingleton())
		{
			const int defaultRuleCount = rule.getDefaultRHSCount();

			//cerr << "default rhs count: " << defaultRuleCount << endl;
			
			switch(defaultRuleCount)
			{
				case 1  :
				{
						tiny_principle::RightHandSide &rhs = rule.getLastRHS();

						if (rhs.hasSelectionAction())
						{
							scanner.error("The default right hand side must be the last right hide side specified.");
							printSymbolString(rule.getLHS()); cerr << ":";
						}
						else
							rhs.getFirstSet()[makeId(scanner.getAnySymbol())] = true;
				}
				//		Fall through
				case 0  :
				{
		 				if (!rule.areRHS_Disjoint())
						{
							scanner.error("conflicting selection sets for the following rule: ", false);
							printSymbolString(rule.getLHS()); cerr << "\n";
							conflict = 1;
						}
				}
				break;
				default :
				{
					if (!scanner.getErrorCount())
					{
						scanner.error(std::string("more than one default rule for: ") +  nt.getName());
				   }
				}
			}

#if 0						
			if (!rule.areRHS_accesible())
			{
				scanner.error("unused righthand side for: ", false);
				printSymbolString(rule.getLHS()); cerr << "\n";
			}
#endif
		}
	}
}

//
//
//
void tiny_symbolProcessor::doComputeInitialFirstSets()
{
	traverse(&tiny_symbolProcessor::computeInitialFirstSet);
}

//
//
//
void tiny_symbolProcessor::computeInitialFirstSet(tiny_rewritableSymbol& nt)
{
	const int ruleCount = nt.getRuleCount();
	
	for(int i = 0; i < ruleCount; i++)
	{
		tiny_principle &rule = nt.getRule(i);
		const int altCount = rule.getAlternativeCount();

		if (!rule.isSingleton())
		{
			for(int j = 0 ; j < altCount; j++)
			{
				tiny_principle::RightHandSide &rhs = rule.getRHS(j);

				if (rhs.hasSelectionAction())
					continue;

				tiny_symbolString &rhsString = rhs.getString();
				const unsigned int index = getDotIndex(&rhsString) + 1;

				if (index == rhsString.length())
					continue;

            if (scanner.getPlaybackSymbol() == rhsString.getSymbol(index))
            {
					rhs.setLeaf(0);
					rhs.setRule(0);
            }				
				else if (isTerminal(rhsString.getSymbol(index)))
				{
					const int id = makeId(rhsString.getSymbol(index));
					rhs.getFirstSet()[id] = true;
					rule.getFirstSet()[id] = true;
					rhs.setLeaf(1);
					rhs.setRule(0);
				}
				else
				{
					rhs.setLeaf(0);
				}
			}
		}
		else	// altCount == 1
		{
			tiny_principle::RightHandSide &rhs = rule.getRHS(0);

			if (rhs.hasSelectionAction())
					continue;

			tiny_symbolString &rhsString = rhs.getString();
			const int index = getDotIndex(&rhsString) + 1;

			if (isTerminal(rhsString.getSymbol(index)) &&
			    (unsigned)index < rhsString.length())
			{
				const int id = makeId(rhsString.getSymbol(index));
				rhs.getFirstSet()[id] = true;
				rule.getFirstSet()[id] = true;
				rhs.setLeaf(1);
				rhs.setRule(0);
			}
		}
	}
}

bool operator<(const tiny_principle& r1,
               const tiny_principle& r2)
{
   int r1len = r1.getLHS().length();
   int r2len = r2.getLHS().length();

   if (isAction(r1.getLHS().getSymbol(r1len - 1)))
      r1len--;

	if (isAction(r2.getLHS().getSymbol(r2len - 1)))
      r2len--;

   if (r1len < r2len)
      return false;
   if (r1len > r2len)
      return true;
   else
      return r1.getId() < r2.getId();
}

//
//
//
void tiny_symbolProcessor::sortSymbolRules(tiny_rewritableSymbol& s)
{
	if (sortRules)
		sort(s.getRules().begin(), s.getRules().end());
}

//
//
//
void tiny_symbolProcessor::processData()
{
   doReserveMemoryForFirstSets();
	
	checkUsage();

	doSortSymbolRules();

	D( cerr << "Contents of the symbolTable" << endl; )
   D( printSymbolTable();	)

	//
	//	Build the graph between left hand and right hand sides
	//
	doComputeConnections();

	D( cerr << "The computed connections" << endl; )
   D( printSymbolTable(); )

	//
	//	Compute atomic first sets
	//
	doComputeInitialFirstSets();

	D( cerr << "The initial firstsets" << endl; )
   D( printSymbolTable();	)

	//
	//	Compute derived first sets
	//
 	computeFirstSets();

   D( cerr << "The compete firstsets" << endl; )
   D( printSymbolTable();	)

	//
	// Compute selection for the grammars;
	//
	doComputeSelectionSets();

	if (conflict)
	        printSymbolTable();

	D( cerr << "the selectionsets" << endl; )
   D( printSymbolTable();	)
}

//
//
//
void tiny_symbolProcessor::computeFirstSets()
{
	do
	{
		didSomething = 0;
		doUpdateFirstSets();
	}
	while(didSomething);

}

//
//
//
void tiny_symbolProcessor::doUpdateFirstSets()
{
	traverse(&tiny_symbolProcessor::updateFirstSets);
}

//
//
//
void tiny_symbolProcessor::updateFirstSets(tiny_rewritableSymbol& nt)
{
	const int ruleCount = nt.getRuleCount();

	for(int i = 0; i < ruleCount; i++)
	{
		tiny_principle &rule = nt.getRule(i);
      //D( printSymbolString(rule.getLHS()); cerr << endl; )

		rule.updateFirstSets(didSomething);
	}
}

//
//
//
void tiny_symbolProcessor::printSymbolTable()
{
	traverse(&tiny_symbolProcessor::printSymbols);
}

//
//
//
void tiny_symbolProcessor::printSymbols(tiny_rewritableSymbol& s)
{
	for (unsigned int i = 0; i < s.getRuleCount(); i++)
	{
		printSymbolString(s.getLHS(i));
		cerr << " : ";
		const unsigned tcount = scanner.getTerminalCount();
		const unsigned altCount = s.getRule(i).getAlternativeCount();
		for(unsigned j = 0; j < altCount; j++)
		{
			tiny_principle::RightHandSide &rhs = s.getRule(i).getRHS(j);
			int sym = rhs.getString().getSymbol(0);

			if (!isAction(sym))
			{
				const vector<bool> &bs = rhs.getFirstSet();
            unsigned anySymbol = makeId(scanner.getAnySymbol());
            unsigned playbackSymbol = makeId(scanner.getPlaybackSymbol());
            bool includesAny = false;
            bool includesPlayback = false;
            int pcount = 0;

				for(unsigned k = 0; k < tcount; ++k)
				{
               if (bs[k])
               {
                  ++pcount;
                  if (k == anySymbol)
                     includesAny = true;
                  else if (k == playbackSymbol)
                     includesPlayback = true;
               }
				}
				
            if (includesAny)
				   cerr << "[$any] ";
            else if (pcount == 1 && includesPlayback)
               ;
            else if (pcount)
				{
   				cerr << "[";

               int printed = 0;
			 		for(unsigned k = 0; k < tcount; k++)
					{
						if (bs[k])
						{
							printSymbol(makeTerminal(k));
                     if (printed < pcount-1)
                     {
                        ++printed;
							   cerr << ", ";
                     }
						}
					}
   				cerr << "]\t ";
				}
			}

			printSymbolString(rhs.getString());

			if (!rhs.getStringToRecord().empty())
			{
				cerr << " records: ";
   			printSymbolString(rhs.getStringToRecord());
			}
			
			if (rhs.getRule())
			{
				cerr << " connects to: ";
				printSymbolString(rhs.getRule()->getLHS());
			}

			cerr << "\n";

			if (j != altCount - 1)
				cerr << "\t| ";
		}

		cerr << "\t;\n\n";
	}
}

//
//
//
void tiny_symbolProcessor::printSymbolString(tiny_symbolString& ss)
{
   unsigned length = ss.length();
	for(unsigned int i = 0; i < length; i++)
	{
		printSymbol(ss.getSymbol(i));
		
		if (i != length - 1)
		   cerr << " ";
   }
}

//
//
//
void tiny_symbolProcessor::printSymbol(int id)
{
	unsigned int newId = makeId(id);

	switch(symbolType(id))
	{
		case TINY_NONTERMINAL_MASK :
		case TINY_GENERIC_MASK :
		{
			const tiny_rewritableSymbol& s = scanner.getNonTerminal(newId);
			cerr << s.getName();
		}
		break;
		case TINY_TERMINAL_MASK :
		{
			const tiny_grammarSymbol& t = scanner.getTerminal(newId);
			cerr << t.getName();
		}
		break;
		case TINY_ACTION_MASK :
		{
			const tiny_action& a = scanner.getAction(newId);
				cerr << a.getName();
		}
		break;
		case VARIABLE_MASK :
		{
				cerr << "@" << makeId(id);
		}
		break;
		case 0 :
			cerr << ".";
		break;
		default:
		{
			char buf[32];
			sprintf(buf,"%d", symbolType(id));
			scanner.fatalError(string("Bad symbol class: ") + buf);
		}
	}
}

//
//
//
void tiny_symbolProcessor::printActions()
{
	traverse(&tiny_symbolProcessor::printAction);
}

//
//
//
void tiny_symbolProcessor::printAction(tiny_action& a)
{
	cerr << a.getName() << " = {% " << a.getString() << " %}\n\n";
}
